(* File: state.ml *)
(* Author: Chad E Brown *)
(* Created: October 2010 *)

open String
open Flags
open Syntax
open Priorityqueue

exception Redeclaration of string
exception FileNotFound of string
exception GenericError of string

let mode : string option ref = ref None
let timeout : int option ref = ref None
let verbosity : int ref = ref 1 (*** Default to 1; 0 should mean print nothing (-verb 0 command line option) ***)
let coq = ref false
let coqchannel : out_channel option ref = ref None
let conjecturename : string ref = ref "claim"
let conjecture : (pretrm * trm) option ref = ref None
let mkproofterm = ref false
let slave = ref false
let coqsig_base : string list ref = ref []
let coqsig_const : (string * stp) list ref = ref []
let coqsig_def : (string * pretrm) list ref = ref []
let coqsig_hyp : (string * pretrm) list ref = ref []
let coqsig_def_trm : (string * trm) list ref = ref []
let coqsig_hyp_trm : (string * trm) list ref = ref []
let name_base : (string,unit) Hashtbl.t = Hashtbl.create 10
let name_tp : (string,stp) Hashtbl.t = Hashtbl.create 100
let name_trm : (string,trm * stp) Hashtbl.t = Hashtbl.create 100
let name_def : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_def_prenorm : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_hyp : (string,trm) Hashtbl.t = Hashtbl.create 100
let coq_used_names : (string,unit) Hashtbl.t = Hashtbl.create 100
let coq_names : (string,string) Hashtbl.t = Hashtbl.create 100
let coq_hyp_names : (string,string) Hashtbl.t = Hashtbl.create 100

let next_fresh_name : int ref = ref 0

let rec get_fresh_name a =
  let x = "__" ^ (string_of_int (!next_fresh_name)) in
  incr next_fresh_name;
  if ((Hashtbl.mem name_base x) || (Hashtbl.mem name_tp x)) then
    get_fresh_name a
  else
    let xa = Name(x,a) in
    begin
      Hashtbl.add name_tp x a;
      Hashtbl.add name_trm x (xa,a);
      (x,xa)
    end

let next_atom = ref 1
let atom_hash : (trm,int) Hashtbl.t = Hashtbl.create 1000
let atom_hash_rev : (int,trm) Hashtbl.t = Hashtbl.create 1000

let max_atom () = (!next_atom) - 1

let new_atom m =
  let a = (!next_atom) in
  Hashtbl.add atom_hash m a;
  Hashtbl.add atom_hash_rev a m;
  if (!verbosity > 4) then if (!verbosity > 8) then (Printf.printf "Atom %d: %s\n" a (trm_str m)) else (Printf.printf "Atom %d\n" a);
  incr next_atom;
  a

let get_atom m =
  try
    Hashtbl.find atom_hash m
  with
  | Not_found ->
      begin
	if (get_bool_flag "SYM_EQ") then (*** Associate m1 = m2 and m2 = m1 with the same integer ***)
	  begin
	    match m with
	    | Ap(Ap(Eq(b),m1),m2) ->
		begin
		  try
		    Hashtbl.find atom_hash (Ap(Ap(Eq(b),m2),m1))
		  with
		  | Not_found -> new_atom m
		end
	    | _ -> new_atom m
	  end
	else
	  new_atom m
      end

let get_literal m =
  match (neg_body m) with
    Some n ->
      begin
	match (neg_body n) with
	| Some _ -> raise (GenericError("DN Bug - get_literal called with dneg " ^ (trm_str m)))
	| None ->
	    (- (get_atom n))
      end
  | None -> (get_atom m)
let literal_to_trm l =
  if (l < 0) then
    (neg (Hashtbl.find atom_hash_rev (- l)))
  else
    (Hashtbl.find atom_hash_rev l)

let initial_branch : trm list ref = ref []
let initial_branch_prenorm : trm list ref = ref []

type clause = int list

let clauses : clause list ref = ref []

type searchoption =
 (*** ProcessProp1(m) For rules requiring one formula in the head (most rules) ***)
    ProcessProp1 of trm
 (*** Mating(plit,nlit,pl,nl) where pl,nl and plit = lit (h pl) and nlit = lit ~(h nl) ***)
  | Mating of int * int * trm list * trm list
 (*** Confront(nplit,nnlit,a,s,t,u,v) where s,t,u,v:a and nplit = lit (s = t) and nnlit is lit (u != v) ***)
  | Confront of int * int * stp * trm * trm * trm * trm
 (*** DefaultElt(a) - create a default element of type a ***)
  | DefaultElt of string
 (*** DefaultEltInst(a) - create a default element of type a ***)
  | DefaultEltInst of string
 (*** NewInst(a,m) - put m:a in the set of instantiations ***)
  | NewInst of stp * trm
 (*** EnumIterDeep(d,a) - Enumerate all terms (using the current constants) of depth d ***)
  | EnumIterDeep of int * stp
 (*** EnumTp(d,ar,a) - work on enumerating types that can be used to choose a polymorphic head (Eq, Forall, Choice) ***)
  | EnumTp of int * stp * stp
 (*** EnumAp(d,Gamma,sigmal,h,c) - ***)
  | EnumAp of (int * stp list * stp list * trm * (trm -> unit))
 (*** Enum(d,Gamma,sigma,c) ***)
  | Enum of (int * stp list * stp * (trm -> unit))
 (*** Filter - use Minisat to filter usable sets ***)
  | Filter of int

type ruleinfo =
 (*** NegPropRule(m) For rules requiring one negative formula in the head, except negated forall ***)
    NegPropRule of trm
 (*** PosPropRule(m) For rules requiring one positive formula in the head, except forall ***)
  | PosPropRule of trm
 (*** InstRule(a,m,n) For instantiating a Ap(Forall(a),m) with n:a ***)
  | InstRule of stp * trm * trm
 (*** FreshRule(a,m,x) For producing a witness x for Ap(Neg,Ap(Forall(a),m)) ***)
  | FreshRule of stp * trm * string
 (*** MatingRule(plit,nlit) For mating rule, only save the literals here ***)
  | MatingRule of int * int
 (*** ConfrontationRule(plit,nlit) For confrontation rule, only save the literals here ***)
  | ConfrontationRule of int * int
 (*** ChoiceRule(eps,pred) ***)
  | ChoiceRule of trm * trm
 (*** ClauseRule(nalllit,int list,trm list) - Universal quantifier, list of lits, list of instantiations to give the clause ***)
  | ClauseRule of int * int list * trm list

type refut =
    NegImpR of trm * trm * refut
  | ImpR of trm * trm * refut * refut
  | FalseR
  | NegReflR
  | NegAllR of stp * trm * string * refut
  | EqFR of stp * stp * trm * trm * refut
  | NegEqFR of stp * stp * trm * trm * refut
  | EqOR of trm * trm * refut * refut
  | NegEqOR of trm * trm * refut * refut
  | SearchR of clause list * (clause -> ruleinfo)
  | AssumptionConflictR

(*** For printing a searchoption - helpful for debugging. - Chad, Oct 2010 ***)
let searchoption_str s =
  match s with
  | ProcessProp1(m) -> ("ProcessProp1(" ^ (trm_str m) ^ ")")
  | Mating(plit,nlit,pl,nl) -> "Mating(" ^ (string_of_int plit) ^ "," ^ (string_of_int nlit) ^ " -- " ^ (String.concat "," (List.map trm_str pl)) ^ " -- " ^ (String.concat "," (List.map trm_str nl)) ^  ")"
  | Confront(plit,nlit,a,s,t,u,v) -> "Confront(" ^ (string_of_int plit) ^ "," ^ (string_of_int nlit) ^ ")"
  | DefaultElt(a) -> "DefaultElt(" ^ a ^ ")"
  | DefaultEltInst(a) -> "DefaultEltInst(" ^ a ^ ")"
  | EnumIterDeep(d,a) -> "EnumIterDeep(" ^ (string_of_int d) ^ "," ^ (stp_str a) ^ ")"
  | EnumTp(d,_,a) -> "EnumTp(" ^ (string_of_int d) ^ "," ^ (stp_str a) ^ ")"
  | EnumAp(d,gamma,sigmal,h,_) -> "EnumAp(" ^ (string_of_int d) ^ "," ^ (String.concat ";" (List.map stp_str gamma)) ^ "," ^ (String.concat ";" (List.map stp_str sigmal)) ^ "," ^ (trm_str h) ^ ")"
  | Enum(d,gamma,sigma,_) -> "Enum(" ^ (string_of_int d) ^ "," ^ (String.concat ";" (List.map stp_str gamma)) ^ "," ^ (stp_str sigma) ^ ")"
  | NewInst(_,m) -> "NewInst(" ^ (trm_str m) ^ ")"
  | Filter(d) -> "Filter(" ^ (string_of_int d) ^ ")"
  | _ -> "OtherSearchOption" (*** to do: the rest if we need them ***)

(*** For printing a ruleinfo - helpful for debugging. - Chad, Oct 2010 ***)
let ruleinfo_str s =
  match s with
  | NegPropRule(m) -> ("NegPropRule(" ^ (trm_str m) ^ ")")
  | PosPropRule(m) -> ("PosPropRule(" ^ (trm_str m) ^ ")")
  | InstRule(a,m,n) -> ("InstRule(" ^ (stp_str a) ^ "," ^ (trm_str m) ^ "," ^ (trm_str n) ^ ")")
  | FreshRule(a,m,x) -> ("FreshRule(" ^ (stp_str a) ^ "," ^ (trm_str m) ^ "," ^ x ^ ")")
  | MatingRule(plit,nlit) -> "MatingRule(" ^ (string_of_int plit) ^ "," ^ (string_of_int nlit) ^ ")"
  | ConfrontationRule(plit,nlit) -> "ConfrontationRule(" ^ (string_of_int plit) ^ "," ^ (string_of_int nlit) ^ ")"
  | ChoiceRule(eps,pred) -> "ChoiceRule(" ^ (trm_str eps) ^ "," ^ (trm_str pred) ^ ")"
  | ClauseRule(nalllit,lits,insts) -> "ClauseRule(" ^ (string_of_int nalllit) ^ "," ^ (String.concat ";" (List.map string_of_int lits)) ^ "," ^ (String.concat ";" (List.map trm_str insts)) ^ ")"

(*** For printing the refutation - helpful for debugging. - Chad, Oct 2010 ***)
let rec print_refut r =
  match r with
  | NegImpR(m,n,r1) ->
      Printf.printf "NegImpR\n  left: %s\n  right: %s\n" (trm_str m) (trm_str n);
      print_refut r1
  | ImpR(m,n,r1,r2) ->
      Printf.printf "ImpR\n  left: %s\n  right: %s\n" (trm_str m) (trm_str n);
      print_refut r1;
      print_refut r2
  | FalseR ->
      Printf.printf "FalseR\n"
  | NegReflR ->
      Printf.printf "NegReflR\n"
  | NegAllR(a,m,x,r) ->
      Printf.printf "NegAllR\n  fresh var:%s\n  stp: %s\n   body: %s\n" x (stp_str a) (trm_str m);
      print_refut r
  | EqFR(a,b,m,n,r) ->
      Printf.printf "EqFR\n  stp: (%s)   ->   (%s)\nleft: %s\n  right: %s\n" (stp_str a) (stp_str b) (trm_str m) (trm_str n);
  | NegEqFR(a,b,m,n,r) ->
      Printf.printf "NegEqFR\n  stp: (%s)   ->   (%s)\nleft: %s\n  right: %s\n" (stp_str a) (stp_str b) (trm_str m) (trm_str n);
  | EqOR(m,n,r1,r2) ->
      Printf.printf "EqOR\n  left: %s\n  right: %s\n" (trm_str m) (trm_str n);
      print_refut r1;
      print_refut r2
  | NegEqOR(m,n,r1,r2) ->
      Printf.printf "NegEqOR\n  left: %s\n  right: %s\n" (trm_str m) (trm_str n);
      print_refut r1;
      print_refut r2
  | SearchR(cl,cr) ->
      Printf.printf "SearchR\n  clauses with search options:\n";
      List.iter
	(fun c ->
	  Printf.printf "%s\n" (String.concat " " (List.map string_of_int c));
	  List.iter
	    (fun z -> (Printf.printf "Atom %d: %s\n" (abs z) (trm_str (Hashtbl.find atom_hash_rev (abs z)))))
	    c;
	  try
	    Printf.printf "associated rule: %s\n" (ruleinfo_str (cr c))
	  with
	  | Not_found ->
	      Printf.printf "no associated rule\n"
	)
	cl

let processed : (trm,unit) Hashtbl.t = Hashtbl.create 100
let clause_ruleinfo : (clause,ruleinfo) Hashtbl.t ref = ref (Hashtbl.create 100)

let insert_assumption_clause c = clauses := (c::!clauses)

let insert_search_clause c r = 
  begin
    clauses := (c::!clauses);
    match r with
    | Some(r) ->
	Hashtbl.add (!clause_ruleinfo) c r;
    | None -> ()
  end

(*** Different Implementations of Fair Priority Queues ***)
module PQueue0 = PriorityQueue0 (struct type t = searchoption end)

module PQueue1 = PriorityQueue1 (struct type t = searchoption end)

let insertWithPriority n o =
  let (ins,dp) =
    begin
      match (get_int_flag "PRIORITY_QUEUE_IMPL") with
      | 0 -> (PQueue0.insertWithPriority,PQueue0.debug_print)
      | 1 -> (PQueue1.insertWithPriority,PQueue1.debug_print)
      | n -> (raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n))))
    end
  in
  begin
    if (!verbosity > 9) then Printf.printf "Inserting option with priority %d: %s\n" n (searchoption_str o);
    ins n o;
    if (!verbosity > 49) then dp searchoption_str
  end

let getNext () =
  let (gn,dp) =
    begin
      match (get_int_flag "PRIORITY_QUEUE_IMPL") with
      | 0 -> (PQueue0.getNext,PQueue0.debug_print)
      | 1 -> (PQueue1.getNext,PQueue1.debug_print)
      | n -> (raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n))))
    end
  in
  if (!verbosity > 49) then (Printf.printf "About to get next option:\n"; dp searchoption_str);
  let o = gn()
  in
  if (!verbosity > 49) then (Printf.printf "After removing option:\n"; dp searchoption_str);
  o

let peekAtNext() =
  begin
    match (get_int_flag "PRIORITY_QUEUE_IMPL") with
    | 0 -> PQueue0.peekAtNext()
    | 1 -> PQueue1.peekAtNext()
    | n -> (raise (GenericError("Invalid value of PRIORITY_QUEUE_IMPL: " ^ (string_of_int n))))
  end

let testPriorityQueue() =
  let v = (!verbosity) in
(***  verbosity := 50; ***)
  for i = 0 to 2000
  do
    let p = (Random.int 10) in
    Printf.printf "Inserting %d with priority %d\n" i p;
    Printf.printf "(ins %d %d)\n" i p;
    insertWithPriority p (Filter i);
    if ((Random.int 3) = 0) then
      begin
	let (q,o1) = peekAtNext() in
	Printf.printf "Peeking at %s of priority %d\n" (searchoption_str o1) q;
	Printf.printf "(unless (equal (getnext) (list %d %d)) (error \"\"))\n" (match o1 with (Filter z) -> z | _ -> 0) q;
	let o2 = getNext() in
	if (not (o1 = o2)) then
	  Printf.printf "Got something different: %s\n" (searchoption_str o2);
      end;
  done;
  verbosity := v

let patoms : (string,int * (trm list)) Hashtbl.t = Hashtbl.create 10
let natoms : (string,int * (trm list)) Hashtbl.t = Hashtbl.create 10

let pchoiceatoms : (stp,int * (trm list)) Hashtbl.t = Hashtbl.create 10
let nchoiceatoms : (stp,int * (trm list)) Hashtbl.t = Hashtbl.create 10

let peqns : (string,int * trm * trm) Hashtbl.t = Hashtbl.create 10
let neqns : (string,int * trm * trm) Hashtbl.t = Hashtbl.create 10

let univpreds : (stp,(int * trm)) Hashtbl.t = Hashtbl.create 10
let instantiations : (stp,(trm,unit) Hashtbl.t) Hashtbl.t = Hashtbl.create 10
let default_elts : (string,trm) Hashtbl.t = Hashtbl.create 10

let set_default_elt aname x =
  Hashtbl.add default_elts aname x

let set_default_elt_if_needed a x =
  match a with
  | Base(aname) ->
      if (not (Hashtbl.mem default_elts aname)) then set_default_elt aname x
  | _ -> ()

let default_elt aname =
  try
    Hashtbl.find default_elts aname
  with
  | Not_found ->
      let a = Base aname in
      let (_,x) = get_fresh_name a in
      begin
	set_default_elt aname x;
	x
      end

let default_elt_p aname =
  Hashtbl.mem default_elts aname

let get_instantiations a =
  try
    let h = Hashtbl.find instantiations a in
    let l = ref [] in
    Hashtbl.iter (fun key _ -> l := (key::!l)) h;
    !l
  with Not_found -> []

let known_instantiation a m =
  try
    let h = Hashtbl.find instantiations a in
    Hashtbl.find h m;
    true
  with Not_found -> false

let add_instantiation a m =
  try
    let h = Hashtbl.find instantiations a in
    Hashtbl.add h m ();
    set_default_elt_if_needed a m
  with Not_found ->
    let h : (trm,unit) Hashtbl.t = Hashtbl.create 10 in
    Hashtbl.add instantiations a h;
    Hashtbl.add h m ();
    set_default_elt_if_needed a m

let choiceopnames : (string,stp) Hashtbl.t = Hashtbl.create 10

let choiceop_axiom m =
  match m with
  | Ap (Forall (Ar (a, Prop)),
	Lam (Ar (_, Prop),
	     Ap (Forall _,
		 Lam (_,
		      Ap (Ap (Imp, Ap (DB (1, Ar (_, Prop)), DB (0, _))),
			  Ap (DB (1, Ar (_, Prop)),
			      Ap (Name (x, Ar (Ar (_, Prop), _)),
				  DB (1, Ar (_, Prop)))))))))
    -> Some(x,a)
  | Ap (Forall (Ar (a, Prop)),
	Lam (Ar (_, Prop),
	     Ap
	       (Ap (Imp,
		    Ap
		      (Ap (Imp,
			   Ap (Forall _,
			       Lam (_,
				    Ap (Ap (Imp, Ap (DB (1, Ar (_, Prop)), DB (0, _))),
					False)))),
		       False)),
		Ap (DB (0, Ar (_, Prop)),
		    Ap (Name (x, Ar (Ar (_, Prop), _)),
			DB (0, Ar (_, Prop)))))))
    -> Some(x,a)
  | _ -> None

let declare_choiceop x a = Hashtbl.add choiceopnames x a

let choiceop m =
  match m with
  | Choice(a) -> Some(a)
  | Name(x,_) ->
      begin
	try
	  Some(Hashtbl.find choiceopnames x)
	with
	| Not_found -> None
      end
  | _ -> None

let filtered : (int,unit) Hashtbl.t = Hashtbl.create 10

type namecategory =
    ChoiceOp of int * int * stp list * trm list (*** (i,n,sigmal,tl) where length of sigmal and tl are n, 0 <= i < n, Name has type (sigmal -> o) -> sigmal[i], and for each j:{0,...,n-1} tl[j] is the jth component of the n-ary choice operator (in particular, tl[i] is this one) ***)
   | DescrOp of int * int * stp list * trm list (*** (i,n,sigmal,tl) where length of sigmal and tl are n, 0 <= i < n, Name has type (sigmal -> o) -> sigmal[i], and for each j:{0,...,n-1} tl[j] is the jth component of the n-ary description operator (in particular, tl[i] is this one) ***)
   | IfThenElse of int * int * stp list (*** (i,n,sigmal) where length of sigmal is n, 0 <= i < n, Name has type o -> sigmal -> sigmal -> sigmal[i] ***)
   | ReflexiveBinary
   | IrreflexiveBinary
   | SymmetricBinary
   | ReflexiveSymmetricBinary
   | IrreflexiveSymmetricBinary

let constrainedName : (string,namecategory) Hashtbl.t = Hashtbl.create 10

let decomposable x =
   try
   let c = Hashtbl.find constrainedName x in (*** Some categorized names are decomposable and some are not ***)
   match c with
   | IfThenElse _ -> false (*** TO DO: Check that n-ary if-then-else need not be decomposable ***)
   (*** TO DO: Should *Binary cases be decomposable? ***)
   | _ -> true
   with Not_found -> true (*** A name is decomposable by default ***)

exception Unsatisfiable of refut option
exception Satisfiable

(*** Set completep to false if I use a mode that makes the search incomplete, so that failure does not imply satisfiability.
     Currently there is no such mode. - Chad, Oct 2010 ***)
let completep = ref true

let get_timeout_default x = match (!timeout) with Some y -> y | None -> x

let st_include_fun : (string -> unit) ref = ref (fun x -> raise (GenericError("Bug when including file " ^ x)))

let coq_init () =
  begin
    Hashtbl.add coq_used_names "as" ();
    Hashtbl.add coq_used_names "at" ();
    Hashtbl.add coq_used_names "cofix" ();
    Hashtbl.add coq_used_names "else" ();
    Hashtbl.add coq_used_names "end" ();
    Hashtbl.add coq_used_names "exists2" ();
    Hashtbl.add coq_used_names "fix" ();
    Hashtbl.add coq_used_names "for" ();
    Hashtbl.add coq_used_names "forall" ();
    Hashtbl.add coq_used_names "fun" ();
    Hashtbl.add coq_used_names "if" ();
    Hashtbl.add coq_used_names "IF" ();
    Hashtbl.add coq_used_names "in" ();
    Hashtbl.add coq_used_names "let" ();
    Hashtbl.add coq_used_names "match" ();
    Hashtbl.add coq_used_names "mod" ();
    Hashtbl.add coq_used_names "Prop" ();
    Hashtbl.add coq_used_names "return" ();
    Hashtbl.add coq_used_names "Set" ();
    Hashtbl.add coq_used_names "then" ();
    Hashtbl.add coq_used_names "Type" ();
    Hashtbl.add coq_used_names "using" ();
    Hashtbl.add coq_used_names "where" ();
    Hashtbl.add coq_used_names "with" ();
    (*** Avoid certain names used by stt in Coq ***)
    Hashtbl.add coq_used_names "SType" ();
    Hashtbl.add coq_used_names "Sar" ();
    Hashtbl.add coq_used_names "o" ();
    Hashtbl.add coq_used_names "Sepsilon" ();
    Hashtbl.add coq_used_names "forall" ();
    Hashtbl.add coq_used_names "exists" ();
    Hashtbl.add coq_used_names "False" ();
    Hashtbl.add coq_used_names "True" ();
    Hashtbl.add coq_used_names "not" ();
    Hashtbl.add coq_used_names "and" ();
    Hashtbl.add coq_used_names "or" ();
    Hashtbl.add coq_used_names "iff" ();
    Hashtbl.add coq_used_names "ex" ();
    Hashtbl.add coq_used_names "eq" ();
    Hashtbl.add coq_used_names "I" ();
    Hashtbl.add coq_used_names "FalseE" ();
    Hashtbl.add coq_used_names "conj" ();
    Hashtbl.add coq_used_names "proj1" ();
    Hashtbl.add coq_used_names "proj2" ();
    Hashtbl.add coq_used_names "and_ind" ();
    Hashtbl.add coq_used_names "or_introl" ();
    Hashtbl.add coq_used_names "or_intror" ();
    Hashtbl.add coq_used_names "or_ind" ();
    Hashtbl.add coq_used_names "ex_intro" ();
    Hashtbl.add coq_used_names "ex_ind" ();
    Hashtbl.add coq_used_names "refl_equal" ();
    Hashtbl.add coq_used_names "eq_ind" ();
    Hashtbl.add coq_used_names "SinhE" ();
    Hashtbl.add coq_used_names "classic" ();
    Hashtbl.add coq_used_names "NNPP" ();
    Hashtbl.add coq_used_names "prop_ext" ();
    Hashtbl.add coq_used_names "functional_extensionality" ();
    Hashtbl.add coq_used_names "Sepsilon_spec" ();
    (*** Other names to avoid ***)
    Hashtbl.add coq_used_names "claim" ();
  end

let print_coqsig c =
  let rec print_coqsig_base l =
    match l with
      (x::r) ->
	print_coqsig_base r;
	Printf.fprintf c "Variable %s:SType.\n" x
    | [] -> ()
  in
  let rec print_coqsig_const l =
    match l with
      ((x,a)::r) ->
	begin
	  print_coqsig_const r;
	  try
	    ignore (List.assoc x (!coqsig_def));
	  with
	  | Not_found ->
	      begin
		try
		  Printf.fprintf c "Variable %s : " x;
		  print_stp_coq c a coq_names false;
		  Printf.fprintf c ".\n"
		with
		| Not_found ->
		    begin
		      if (c != stdout) then close_out c;
		      raise (GenericError("A Satallax bug caused a problem creating the Coq file."))
		    end
	      end
	end
    | [] -> ()
  in
  let rec print_coqsig_def l =
    match l with
      ((x,a)::r) ->
	begin
	  print_coqsig_def r;
	  try
	    let m = List.assoc x (!coqsig_def) in
	    begin
	      Printf.fprintf c "Definition %s : " x;
	      print_stp_coq c a coq_names false;
	      Printf.fprintf c "\n := ";
	      print_pretrm_coq c m coq_names coq_used_names (-1) (-1);
	      Printf.fprintf c ".\n"
	    end
	  with
	  | Not_found -> ()
	end
    | [] -> ()
  in
  let rec print_coqsig_hyp l =
    match l with
      ((x,m)::r) ->
	begin
	  try
	    print_coqsig_hyp r;
	    Printf.fprintf c "Hypothesis %s : " x;
	    print_pretrm_coq c m coq_names coq_used_names (-1) (-1);
	    Printf.fprintf c ".\n"
	  with
	  | Not_found ->
	      begin
		if (c != stdout) then close_out c;
		raise (GenericError("A Satallax bug caused a problem creating the Coq file."))
	      end
	end
    | [] -> ()
  in
  begin
    Printf.fprintf c "Add LoadPath \"%s/coq\".\n" Config.satallaxdir;
    Printf.fprintf c "Require Import stttab.\n";
    print_coqsig_base !coqsig_base;
    print_coqsig_const !coqsig_const;
    print_coqsig_def !coqsig_const;
    print_coqsig_hyp !coqsig_hyp;
    match (!conjecture) with
    | Some (m,_) ->
	begin
	  Printf.fprintf c "Theorem %s : " (Hashtbl.find coq_hyp_names (!conjecturename));
	  print_pretrm_coq c m coq_names coq_used_names (-1) (-1);
	  Printf.fprintf c ".\n"
	end
    | None -> Printf.fprintf c "Theorem claim : False.\n"
  end

let declare_base_type a =
  if (!coq) then
    begin 
      let y = coqify_name a coq_names coq_used_names in
      coqsig_base := (y::!coqsig_base)
    end;
  Hashtbl.add name_base a ();
  if (get_bool_flag "CHOICE_AS_DEFAULT") then
    set_default_elt a (norm name_def (ap(Choice (Base a),Lam(Base a,False))))

let st_to_stp m =
  begin
    try
      to_stp m
    with
    | DeclareInd ->
	begin (*** Declare the base type i the first time it's seen.  Use a different name if an 'i' has already been used. ***)
	  declare_base_type "$i";
	  to_stp m
	end
  end

let st_to_trm_given_stp m tp =
  begin
    try
      let (n,_) = to_trm name_trm [] m (Some tp) in n
    with
    | DeclareInd ->
	begin (*** Declare the base type i the first time it's seen.  Use a different name if an 'i' has already been used. ***)
	  declare_base_type "$i";
	  let (n,_) = to_trm name_trm [] m (Some tp) in n
	end
  end

let st_to_trm m =
  begin
    try
      to_trm name_trm [] m None
    with
    | DeclareInd ->
	begin (*** Declare the base type i the first time it's seen.  Use a different name if an 'i' has already been used. ***)
	  declare_base_type "$i";
	  to_trm name_trm [] m None
	end
  end

let declare_typed_constant (name:string) (role:string) (m:pretrm) =
  begin
    if (!verbosity > 4) then Printf.printf "declare_typed_constant %s %s\n%s\n" name role (pretrm_str m);
    match m with
      POf(PName(x),m) ->
	begin
	  match m with
	    PType ->
	      if (Hashtbl.mem name_base x) then raise (Redeclaration x);
	      if (Hashtbl.mem name_tp x) then raise (Redeclaration x);
	      declare_base_type x
	  | _ ->
	      let tp = (st_to_stp m) in
	      if (Hashtbl.mem name_base x) then raise (Redeclaration x);
	      if (Hashtbl.mem name_tp x) then raise (Redeclaration x);
	      if (!coq) then
		begin
		  let y = coqify_name x coq_names coq_used_names in
		  coqsig_const := (y,tp)::!coqsig_const
		end;
	      Hashtbl.add name_tp x tp;
	      Hashtbl.add name_trm x (Name(x,tp),tp)
	end
    | _ -> raise (GenericError ("Incorrect format for type declaration " ^ name))
  end
    
let declare_definition_real x m =
  if (Hashtbl.mem name_base x) then raise (Redeclaration x);
  if (Hashtbl.mem name_def x) then raise (Redeclaration x);
  try
    let tp = Hashtbl.find name_tp x in
    let tm = st_to_trm_given_stp m tp in
    if (!coq) then
      begin (*** The name was coqified when the type was declared. ***)
	try
	  let y = Hashtbl.find coq_names x in
	  coqsig_def := (y,m)::!coqsig_def;
	  coqsig_def_trm := (y,tm)::!coqsig_def_trm
	with
	| Not_found -> raise (GenericError("Could not find Coq version of name " ^ x))
      end;
    Hashtbl.add name_def_prenorm x tm;
    Hashtbl.add name_def x (norm name_def (logicnorm tm))
  with
  | Not_found ->
      begin
	let (tm,tp) = st_to_trm m in
	if (!coq) then
	  begin
	    let y = coqify_name x coq_names coq_used_names in
	    coqsig_const := (y,tp)::!coqsig_const;
	    coqsig_def := (y,m)::!coqsig_def;
	    coqsig_def_trm := (y,tm)::!coqsig_def_trm
	  end;
	Hashtbl.add name_tp x tp;
	Hashtbl.add name_trm x (Name(x,tp),tp);
	Hashtbl.add name_def_prenorm x tm;
	Hashtbl.add name_def x (norm name_def (logicnorm tm))
      end
	
let declare_definition (name:string) (role:string) (m:pretrm) =
  begin
    if !verbosity > 4 then Printf.printf "declare_definition %s %s\n" name role;
    match m with
      PDef(PName(x),m) ->
	declare_definition_real x m
    | PAp(PAp(PEq,PName(x)),m) ->
	declare_definition_real x m
    | _ -> raise (GenericError ("Incorrect format for definition " ^ name))
  end

let declare_thf_logic_formula (name:string) (role:string) (m:pretrm) =
  begin
    if !verbosity > 4 then Printf.printf "declare_thf_logic_formula %s %s\n" name role;
    if ((role = "axiom") || (role = "hypothesis") || (role = "lemma")) then
      begin
	if (Hashtbl.mem name_hyp name) then raise (Redeclaration name);
	let tm = st_to_trm_given_stp m Prop in
	if (!coq) then
	  begin
	    let y = coqify_name name coq_hyp_names coq_used_names in
	    coqsig_hyp := ((y,m)::!coqsig_hyp);
	    coqsig_hyp_trm := ((y,tm)::!coqsig_hyp_trm)
	  end;
	let tmn = norm name_def (logicnorm tm) in
	Hashtbl.add name_hyp name tmn;
	initial_branch_prenorm := (tm::(!initial_branch_prenorm));
	initial_branch := (tmn::(!initial_branch))
      end
    else if ((role = "conjecture") || (role = "theorem")) then
      begin
	match (!conjecture) with
	| Some _ -> raise (GenericError "Problem file has more than one conjecture.")
	| None ->
	    let tm = st_to_trm_given_stp m Prop in
	    begin
	      conjecturename := name;
	      if (name != "claim") then
		ignore (coqify_name name coq_hyp_names coq_used_names);
	      conjecture := Some (m,tm);
	      let ntm = Ap(Neg,tm) in
	      let ntmn = norm name_def (logicnorm ntm) in
	      initial_branch_prenorm := (ntm::(!initial_branch_prenorm));
	      initial_branch := (ntmn::(!initial_branch))	      
	    end
      end
    else
      raise (GenericError ("Unknown role " ^ role))
  end

(*** Code for enumeration of types and terms - Dec 10, 2010 - Chad ***)
let enum_started = ref false
let enum_of_started_ : (stp,unit) Hashtbl.t = Hashtbl.create 5
let enum_of_started a =
  Hashtbl.mem enum_of_started_ a
let enum_of_start a =
  Hashtbl.add enum_of_started_ a ()
let type_continuations_rtp : (stp option,(stp -> int -> unit)) Hashtbl.t = Hashtbl.create 5
let term_continuations_rtp : (stp,(stp list * trm * int -> unit)) Hashtbl.t = Hashtbl.create 5
let usableTypes_rtp : (stp,(stp * int)) Hashtbl.t = Hashtbl.create 5
let usableTypes : (stp * stp * int) list ref = ref []
let usableHeads_rtp : (stp,(stp list * trm * int)) Hashtbl.t = Hashtbl.create 5

let new_type_continuation_rtp ar f =
  Hashtbl.add type_continuations_rtp (Some ar) f

let new_type_continuation f =
  Hashtbl.add type_continuations_rtp None f

let iter_type_continuations_rtp ar a d =
  List.iter (fun f -> f a d) (Hashtbl.find_all type_continuations_rtp (Some ar))

let iter_type_continuations a d =
  List.iter (fun f -> f a d) (Hashtbl.find_all type_continuations_rtp None)

let new_term_continuation_rtp ar f =
  Hashtbl.add term_continuations_rtp ar f

let iter_term_continuations_rtp ar sigmal m p =
  List.iter (fun f -> f (sigmal,m,p)) (Hashtbl.find_all term_continuations_rtp ar)

let new_usable_type_rtp ar a d =
  Hashtbl.add usableTypes_rtp ar (a,d);
  usableTypes := ((ar,a,d)::(!usableTypes))

let usable_types_rtp ar = Hashtbl.find_all usableTypes_rtp ar

let usable_types () = !usableTypes

let new_usable_head_rtp ar sigmal m n = Hashtbl.add usableHeads_rtp ar (sigmal,m,n)

let usable_heads_rtp ar = Hashtbl.find_all usableHeads_rtp ar

(*** Pattern Clauses ***)

(*** search init ***)
let search_init () =
  (*** Add initial instantiations: true and false for o ***)
  add_instantiation Prop False;
  add_instantiation Prop (neg False)
  
(*** reset search ***)
let reset_search () =
  begin
    clauses := [];
    clause_ruleinfo := Hashtbl.create 100;
    PQueue0.reset();
    PQueue1.reset();
    Hashtbl.clear patoms;
    Hashtbl.clear natoms;
    Hashtbl.clear pchoiceatoms;
    Hashtbl.clear nchoiceatoms;
    Hashtbl.clear peqns;
    Hashtbl.clear neqns;
    Hashtbl.clear univpreds;
    Hashtbl.clear instantiations;
    Hashtbl.clear processed;
    enum_started := false;
    Hashtbl.clear enum_of_started_;
    Hashtbl.clear type_continuations_rtp;
    Hashtbl.clear term_continuations_rtp;
    Hashtbl.clear usableTypes_rtp;
    usableTypes := [];
    Hashtbl.clear usableHeads_rtp;
    Hashtbl.clear choiceopnames;
    Hashtbl.clear filtered;
    search_init()
  end

